!---------------------------------------------------------------------------------------
! FLEXINVERT: mod_fluxes
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_fluxes
!!
!! Purpose:    Module for constructing the fluxes data structure.
!!
!---------------------------------------------------------------------------------------


module mod_fluxes

  use mod_settings
  use mod_var

  implicit none
  private

  public :: fluxes_t, alloc_fluxes, dealloc_fluxes

  type :: fluxes_t
 
    real, dimension(:,:,:), allocatable     :: flxnee      ! global NEE fluxes monthly
    real, dimension(:,:,:), allocatable     :: flxff       ! global fossil fuel fluxes monthly
    real, dimension(:,:,:), allocatable     :: flxocn      ! global ocean fluxes monthly
    real, dimension(:,:,:), allocatable     :: flx         ! global fluxes of GHG tracer
    real, dimension(:,:,:), allocatable     :: flxnee_nest ! regional NEE fluxes on flexpart timestamp
    real, dimension(:,:,:), allocatable     :: flxff_nest  ! regional fossil fuel fluxes on flexpart timestamp
    real, dimension(:,:,:), allocatable     :: flxocn_nest ! regional ocean fluxes (only if using nest)
    real, dimension(:,:,:), allocatable     :: flx_nest    ! regional fluxes of GHG tracer (only if using nest)
    real, dimension(:,:,:), allocatable     :: flx_box     ! regional fluxes of tracer on aggregated grid 
    real(kind=8), dimension(:), allocatable :: time        ! time stamp of global fluxes (julian days)
    real(kind=8), dimension(:), allocatable :: timefp      ! time stamp of footprints for whole inversion period

  end type

  contains

  ! --------------------------------------------------
  ! alloc_fluxes
  ! --------------------------------------------------

  subroutine alloc_fluxes(config, fluxes)

    type (config_t),      intent(in)     :: config
    type (fluxes_t),      intent(in out) :: fluxes
    integer                              :: ierr 

    if (config%spec.eq.'co2') then
      ! CO2 species
      allocate( fluxes%flxnee(nxgrid,nygrid,ntstate), stat = ierr )
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR alloc_fluxes: not enough memory'
        stop
      endif
      allocate( fluxes%flxff(nxgrid,nygrid,ntstate), stat = ierr )
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR alloc_fluxes: not enough memory'
        stop
      endif
      allocate( fluxes%flxocn(nxgrid,nygrid,ntstate), stat = ierr )
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR alloc_fluxes: not enough memory'
        stop
      endif
      allocate( fluxes%time(ntstate), stat = ierr )
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR alloc_fluxes: not enough memory'
        stop
      endif
      allocate( fluxes%timefp(nday*ndgrid), stat = ierr )
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR alloc_fluxes: not enough memory'
        stop
      endif
      allocate( fluxes%flxff_nest(nxregrid,nyregrid,nday*ndgrid), stat = ierr )
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR alloc_fluxes: not enough memory'
        stop
      endif
      allocate( fluxes%flxnee_nest(nxregrid,nyregrid,nday*ndgrid), stat = ierr )
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR alloc_fluxes: not enough memory'
        stop
      endif
      if ( config%nested ) then
        allocate( fluxes%flxocn_nest(nxregrid,nyregrid,ntstate), stat = ierr )
        if ( ierr.ne.0 ) then
          write(logid,*) 'ERROR alloc_fluxes: not enough memory'
          stop
        endif
      endif
    else
      ! GHG species
      allocate( fluxes%flx(nxgrid,nygrid,ntstate), stat = ierr )
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR alloc_fluxes: not enough memory'
        stop
      endif
      allocate( fluxes%flx_box(nbox,ntstate,1), stat = ierr )
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR alloc_fluxes: not enough memory'
        stop
      endif
      allocate( fluxes%time(ntstate), stat = ierr )
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR alloc_fluxes: not enough memory'
        stop
      endif
      if ( config%nested ) then
        allocate( fluxes%flx_nest(nxregrid,nyregrid,ntstate), stat = ierr )
        if ( ierr.ne.0 ) stop 'ERROR alloc_fluxes: not enough memory'
      endif
    endif

  end subroutine alloc_fluxes

  ! --------------------------------------------------
  ! dealloc_fluxes
  ! --------------------------------------------------

  subroutine dealloc_fluxes(fluxes)

    type (fluxes_t),      intent(in out) :: fluxes

    if ( allocated(fluxes%flxnee) )       deallocate(fluxes%flxnee)
    if ( allocated(fluxes%flxff)  )       deallocate(fluxes%flxff)
    if ( allocated(fluxes%flxocn) )       deallocate(fluxes%flxocn)
    if ( allocated(fluxes%flxff_nest) )   deallocate(fluxes%flxff_nest)
    if ( allocated(fluxes%flxnee_nest) )  deallocate(fluxes%flxnee_nest)
    if ( allocated(fluxes%flxocn_nest) )  deallocate(fluxes%flxocn_nest)
    if ( allocated(fluxes%flx_nest) )     deallocate(fluxes%flx_nest)
    if ( allocated(fluxes%flx_box) )      deallocate(fluxes%flx_box)
    if ( allocated(fluxes%time) )         deallocate(fluxes%time)
    if ( allocated(fluxes%timefp) )       deallocate(fluxes%timefp)
    if ( allocated(fluxes%flx) )          deallocate(fluxes%flx)

  end subroutine dealloc_fluxes

  ! --------------------------------------------------

end module mod_fluxes




