!---------------------------------------------------------------------------------------
! FLEXINVERT: mod_regions
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_regions
!!
!! Purpose:    Module for defining the mapping from the flexpart to the inversion grid
!!             as well as the areas and local times of the grid. 
!!
!! Details:    The inversion grid is defined as a mapping from the flexpart grid, where
!!             mapping operator is the matrix nbox_xy. The mapping can be based on
!!             either an existing file or the full resolution of the flexpart grid.
!!             In nbox_xy, regions are numbered from 1 to n_land for land regions and 
!!             from -1 to -n_water for ocean regions.
!!       
!---------------------------------------------------------------------------------------


module mod_regions

  use mod_settings
  use mod_var
  use mod_tools
  use netcdf
  use mod_save, only : check 

  implicit none
  private

  public :: get_regions, read_regions, grid_boxes, grid_operator

  contains

  ! --------------------------------------------------
  ! get_regions
  ! --------------------------------------------------
  !> get_regions
  !!
  !! Purpose:  Calculates the area of each of the
  !!           aggregated grid cells, the local
  !!           time at the centre of each grid cell,
  !!           and the land-sea mask for the grid.
  !!
  ! --------------------------------------------------

  subroutine get_regions(files, config)  

    implicit none

    type (files_t),  intent(in)                :: files
    type (config_t), intent(in)                :: config

    integer                                    :: ix, jy     
    integer, dimension(nxregrid,nyregrid)      :: hloc_xy
    real(kind=8), dimension(nxregrid*nyregrid) :: work
    integer, dimension(nxregrid*nyregrid)      :: ind
    integer                                    :: hh
    integer, dimension(:), allocatable         :: cnt
    character(len=max_path_len)                :: rowfmt
    integer                                    :: n, i, ierr
    real                                       :: area

    ! define region boxes
    ! -------------------
 
    if ( config%regions ) then
      ! aggregation of grid using regions file
      write(logid,*) 'Using spatial aggregation...'
      call read_regions(files, config)
    else
      ! no aggregation of grid
      write(logid,*) 'Using no spatial aggregation...'
      call grid_boxes(config)
    endif 
  
    ! calculate mapping operator to regions
    ! not used any more - RLT 06/21
!    call grid_operator(files)

    ! area of each box
    ! ----------------

    ! lat and lon dimensions of region

    allocate( reg_lon(nxregrid) )
    allocate( reg_lat(nyregrid) )

    do ix = 1, nxregrid
      reg_lon(ix) = rllx + (ix-1)*rdx
    end do
    do jy = 1, nyregrid
      reg_lat(jy) = rlly + (jy-1)*rdy
    end do

    allocate( area_box(nbox) )
    area_box(:) = 0.
    if ( nocn.gt.0 ) then
      allocate( area_ocn(nocn) )
      area_ocn(:) = 0.
    endif

    do jy = 1, nyregrid
      area = grid_area(reg_lat(jy)+0.5*rdy, rdy, rdx)
      do ix = 1, nxregrid
        if ( nbox_xy(ix,jy).gt.0 ) then
          area_box(nbox_xy(ix,jy)) = area_box(nbox_xy(ix,jy)) + area 
        else if ( nocn.gt.0 ) then
          area_ocn(-1*nbox_xy(ix,jy)) = area_ocn(-1*nbox_xy(ix,jy)) + area 
        endif
      end do
    end do

    ! local time adjustment for boxes
    ! -------------------------------

    do ix = 1, nxregrid
      hh = floor(reg_lon(ix)*24./360.)
      hloc_xy(ix,:) = hh
    end do

    allocate( hloc(nbox) )
    allocate( cnt(nbox) )
    hloc(:) = 0
    cnt(:) = 0

    ! local time of region is median time of grid cells in box
    work(:) = 0d0
    do n = 1, nbox
      i = 0
      do jy = 1, nyregrid
        do ix = 1, nxregrid
          if( nbox_xy(ix,jy).eq.n ) then
            i = i + 1
            work(i) = dble(hloc_xy(ix,jy))
          endif
        end do
      end do
      call sort(i, work(1:i), ind(1:i))
      hloc(n) = sngl(work(i/2+1))
    end do

    deallocate(cnt)

    ! save nbox_xy and lsm_box
    ! ------------------------

    open(100,file=trim(files%path_output)//'nbox_xy.txt',action='write',status='replace',iostat=ierr)
    write(rowfmt,'(A,I6,A)') '(',nyregrid,'(I5,1X))'
    do ix = 1, nxregrid
      write(100,fmt=rowfmt) nbox_xy(ix,:)
    end do
    close(100)

    open(100,file=trim(files%path_output)//'lsm_box.txt',action='write',status='replace',iostat=ierr)
    write(rowfmt,'(A,I6,A)') '(',nyregrid,'(I5,1X))'
    do ix = 1, nxregrid
      write(100,fmt=rowfmt) lsm_box(ix,:)
    end do
    close(100)

    open(100,file=trim(files%path_output)//'area_box.txt',action='write',status='replace',iostat=ierr)
    do n = 1, nbox
      write(100,fmt='(E11.4)') area_box(n)
    end do
    close(100)

    open(100,file=trim(files%path_output)//'hloc_box.txt',action='write',status='replace',iostat=ierr)
    do n = 1, nbox
      write(100,fmt='(F4.1)') hloc(n)
    end do
    close(100)

  end subroutine get_regions

  ! --------------------------------------------------
  ! read_regions
  ! --------------------------------------------------
  !> read_regions
  !! 
  !! Purpose:  Reads the regions definition file and
  !!           extracts the regions needed for the
  !!           inversion domain. Calculates the 
  !!           mapping operator.
  !!
  ! --------------------------------------------------

  subroutine read_regions(files, config)                

    implicit none

    type (files_t),  intent(in)           :: files
    type (config_t), intent(in)           :: config

    character(len=max_path_len)           :: filename
    character(len=max_name_len)           :: varname, lonname, latname
    logical                               :: lexist
    integer                               :: ncid, lon_dimid, lat_dimid
    integer                               :: lonid, latid, varid
    integer                               :: nlon, nlat
    real, dimension(:), allocatable       :: gbl_rlon
    real, dimension(:), allocatable       :: gbl_rlat
    integer, dimension(:,:), allocatable  :: gbl_reg
    integer                               :: ix1, ix2, jy1, jy2
    integer                               :: n_water, n_invert, n

    filename = files%file_regions
    varname = files%varname_regs
    lonname = files%lonname_regs
    latname = files%latname_regs

    inquire(file=trim(filename),exist=lexist)
    if (.not.lexist) then
      write(logid,*) 'ERROR: cannot find '//trim(filename)
      stop
    endif

    ! read regions file
    ! -----------------

    write(logid,*) 'Reading file: '//trim(filename)

    call check( nf90_open(trim(filename),nf90_NOWRITE,ncid) )
    call check( nf90_inq_dimid(ncid,trim(lonname),lon_dimid) )
    call check( nf90_inq_dimid(ncid,trim(latname),lat_dimid) )

    call check( nf90_inquire_dimension(ncid,lon_dimid,len=nlon) )
    call check( nf90_inquire_dimension(ncid,lat_dimid,len=nlat) )

    if ( .not.allocated(gbl_rlon) ) allocate( gbl_rlon(nlon) )
    if ( .not.allocated(gbl_rlat) ) allocate( gbl_rlat(nlat) )
    if ( .not.allocated(gbl_reg) )  allocate( gbl_reg(nlon,nlat) )

    call check( nf90_inq_varid(ncid,trim(lonname),lonid) )
    call check( nf90_get_var(ncid,lonid,gbl_rlon) )
    call check( nf90_inq_varid(ncid,trim(latname),latid) )
    call check( nf90_get_var(ncid,latid,gbl_rlat) )
    call check( nf90_inq_varid(ncid,trim(varname),varid) )
    call check( nf90_get_var(ncid,varid,gbl_reg) )

    call check( nf90_close(ncid) )

    ! regions for inversion domain
    ! ----------------------------

    ix1 = minloc(gbl_rlon,dim=1,mask=abs(gbl_rlon-rllx-0.5*rdx).eq.minval(abs(gbl_rlon-rllx-0.5*rdx)))
    ix2 = minloc(gbl_rlon,dim=1,mask=abs(gbl_rlon-rurx+0.5*rdx).eq.minval(abs(gbl_rlon-rurx+0.5*rdx)))
    jy1 = minloc(gbl_rlat,dim=1,mask=abs(gbl_rlat-rlly-0.5*rdy).eq.minval(abs(gbl_rlat-rlly-0.5*rdy)))
    jy2 = minloc(gbl_rlat,dim=1,mask=abs(gbl_rlat-rury+0.5*rdy).eq.minval(abs(gbl_rlat-rury+0.5*rdy)))
    write(logid,*) 'read_regions: ix1, ix2, jy1, jy2 = ',ix1,ix2,jy1,jy2

    if ( (nlon.lt.(ix1+nxregrid-1)).or.(nlat.lt.(jy1+nyregrid-1)).or. &
          (ix1.eq.0).or.(ix2.eq.0).or.(jy1.eq.0).or.(jy2.eq.0) ) then
      write(logid,*) 'ERROR: incompatible dimensions in '//trim(filename)
      stop
    endif

    n_water = 0
    n_invert = 0
    do n = minval(gbl_reg(ix1:ix2,jy1:jy2)), maxval(gbl_reg(ix1:ix2,jy1:jy2))
      if ( any(gbl_reg(ix1:ix2,jy1:jy2).eq.n) ) then
         if ( (n.lt.0).and.(.not.config%inc_ocean) ) then
           n_water = n_water - 1
           where( gbl_reg(ix1:ix2,jy1:jy2).eq.n ) nbox_xy = n_water
         else
           n_invert = n_invert + 1
           where( gbl_reg(ix1:ix2,jy1:jy2).eq.n ) nbox_xy = n_invert
         endif  
         if ( n.lt.0 ) then
           where( gbl_reg(ix1:ix2,jy1:jy2).eq.n ) lsm_box = 0
         else
           where( gbl_reg(ix1:ix2,jy1:jy2).eq.n ) lsm_box = 1
         endif
      endif
    end do
    nbox = maxval(nbox_xy)

    if (n_water.ne.0) then
      nocn = abs(minval(nbox_xy)) 
    else 
      nocn = 0
    endif
 
    write(logid,*) 'Number of regions to invert (nbox):',nbox
    if ( .not.config%inc_ocean ) write(logid,*) 'Number of ocean regions (nocn):',nocn

    deallocate(gbl_rlon)
    deallocate(gbl_rlat)
    deallocate(gbl_reg)

  end subroutine read_regions

  ! --------------------------------------------------
  ! grid_boxes
  ! --------------------------------------------------
  !> grid_boxes
  !!
  !! Purpose:  Calculates the mapping operator using
  !!           the flexpart resolution. 
  !!
  ! --------------------------------------------------

  subroutine grid_boxes(config)

    implicit none

    type (config_t), intent(in) :: config

    integer                     :: ix, jy 
    integer                     :: n_water, n_invert

    ! calculate nbox_xy
    n_water = 0
    n_invert = 0
    do jy = 1, nyregrid
      do ix = 1, nxregrid
        if ( (.not.config%inc_ocean).and.(lsm(ix,jy).lt.0.01) ) then
          n_water = n_water - 1
          nbox_xy(ix,jy) = n_water
        else
          n_invert = n_invert + 1
          nbox_xy(ix,jy) = n_invert
        endif 
        if ( lsm(ix,jy).lt.0.01 ) then
          lsm_box(ix,jy) = 0
        else 
          lsm_box(ix,jy) = 1
        endif
      end do
    end do
    nbox = maxval(nbox_xy)

    if (n_water.ne.0) then 
      nocn = abs(minval(nbox_xy)) 
    else 
      nocn = 0
    endif

    write(logid,*) 'Number of regions (nbox):',nbox
    if ( .not.config%inc_ocean ) write(logid,*) 'Number of ocean regions (nocn):',nocn

  end subroutine grid_boxes

  ! --------------------------------------------------
  ! grid_operator
  ! --------------------------------------------------
  !> grid_operator
  !!
  !! Purpose:  Calculates a matrix mapping operator 
  !!           from nbox_xy
  !!
  ! --------------------------------------------------

  subroutine grid_operator(files)

    implicit none
 
    type (files_t),  intent(in)        :: files
    real, dimension(nxregrid*nyregrid) :: vec
    integer                            :: i, j, ierr

    allocate( gridoper(nbox,nxregrid*nyregrid), stat = ierr )
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR grid_operator: not enough memory'
      stop
    endif
    gridoper(:,:) = 0.

    ! calculate gridoper for positive values of nbox_xy (i.e. those inverted)
    ! vectorise nbox_xy
    do j = 1, nyregrid
      do i = 1, nxregrid
        vec((j-1)*nxregrid+i) = nbox_xy(i,j)
      end do
    end do
    ! calculate matrix operator for land
    do i = 1, nbox
      where(vec.eq.i) gridoper(i,:) = 1.
    end do

   open(100,file=trim(files%path_output)//'gridoper.txt',action='write',status='replace',iostat=ierr)
   do i = 1, nbox
     write(100,*) gridoper(i,:)
   end do
   close(100)

   ! to do: sparse matrix form of gridoper to speed up matrix calculations   
 
  end subroutine grid_operator

  ! --------------------------------------------------

end module mod_regions





