!---------------------------------------------------------------------------------------
! FLEXINVERT: read_bg
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> read_bg
!!
!! Purpose:    Reads the "best guess" fluxes from file and calculates the state vector.
!!
!---------------------------------------------------------------------------------------

subroutine read_bg(files, config, states)

  use mod_var
  use mod_settings
  use mod_states
  use mod_tools,   only : grid_area
  use mod_save,    only : check 
  use netcdf

  implicit none

  type (files_t),             intent (in)     :: files
  type (config_t),            intent (in)     :: config
  type (states_t),            intent (in out) :: states

  character(len=max_path_len)                 :: filename
  character(len=*), parameter                 :: lonname='longitude'
  character(len=*), parameter                 :: latname='latitude'
  character(len=*), parameter                 :: timename='time'
  character(len=*), parameter                 :: ndtname = 'hours'
  character(len=*), parameter                 :: xpos_name='xpos'
  integer                                     :: ncid, lon_dimid, lat_dimid, time_dimid, ndt_dimid, varid
  real                                        :: area
  real, dimension(:,:,:,:), allocatable       :: xpos
  real, dimension(:,:,:), allocatable         :: xpos_box
  integer                                     :: nlon, nlat, nt, nh
  integer                                     :: ix, jy, n

  filename = files%file_bg

  if ( config%spec.eq.'co2' ) then

    call check( nf90_open(trim(filename),nf90_NOWRITE,ncid) )
    call check( nf90_inq_dimid(ncid,lonname,lon_dimid) )
    call check( nf90_inq_dimid(ncid,latname,lat_dimid) )
    call check( nf90_inq_dimid(ncid,timename,time_dimid) )
    call check( nf90_inq_dimid(ncid,ndtname,ndt_dimid) )
    call check( nf90_inquire_dimension(ncid,lon_dimid,len=nlon) )
    call check( nf90_inquire_dimension(ncid,lat_dimid,len=nlat) )
    call check( nf90_inquire_dimension(ncid,time_dimid,len=nt) )
    call check( nf90_inquire_dimension(ncid,ndt_dimid,len=nh) )

    ! check dimensions
    if ( (nlon.ne.nxregrid).or.(nlat.ne.nyregrid).or.(nt.ne.ntstate).or.(nh.ne.ndt) ) then
      write(logid,*) 'ERROR: incompatible dimensions in '//trim(filename)
      stop
    endif

    allocate( xpos(nlon,nlat,nt,nh) )
    call check( nf90_inq_varid(ncid,xpos_name,varid) )
    call check( nf90_get_var(ncid,varid,xpos) )
    call check( nf90_close(ncid) )

  else if ( config%spec.eq.'ghg' ) then

    call check( nf90_open(trim(filename),nf90_NOWRITE,ncid) )
    call check( nf90_inq_dimid(ncid,lonname,lon_dimid) )
    call check( nf90_inq_dimid(ncid,latname,lat_dimid) )
    call check( nf90_inq_dimid(ncid,timename,time_dimid) )
    call check( nf90_inquire_dimension(ncid,lon_dimid,len=nlon) )
    call check( nf90_inquire_dimension(ncid,lat_dimid,len=nlat) )
    call check( nf90_inquire_dimension(ncid,time_dimid,len=nt) )

    ! check dimensions
    if ( (nlon.ne.nxregrid).or.(nlat.ne.nyregrid).or.(nt.ne.ntstate) ) then
      write(logid,*) 'ERROR: incompatible dimensions in '//trim(filename)
      stop
    endif

    allocate( xpos(nlon,nlat,nt,1) )
    call check( nf90_inq_varid(ncid,xpos_name,varid) )
    call check( nf90_get_var(ncid,varid,xpos(:,:,:,1)) )
    call check( nf90_close(ncid) )

  endif

  ! numerical scaling
  xpos = xpos*numscale

  ! aggregate to boxes
  if ( config%spec.eq.'co2' ) then
    allocate( xpos_box(nbox,ntstate,ndt) )
    xpos_box(:,:,:) = 0.
  else if ( config%spec.eq.'ghg' ) then
    allocate( xpos_box(nbox,ntstate,1) )
    xpos_box(:,:,:) = 0.
    ndt = 1
  endif
  do n = 1, ntstate
    do jy = 1, nyregrid
      area = grid_area(reg_lat(jy)+0.5*rdy, rdy, rdx)
      do ix = 1, nxregrid
        if ( nbox_xy(ix,jy).gt.0 ) then  
          do nt = 1, ndt
            xpos_box(nbox_xy(ix,jy),n,nt) = xpos_box(nbox_xy(ix,jy),n,nt) + xpos(ix,jy,n,nt) * area
          end do
        endif
      end do
    end do
    do nt = 1, ndt
      xpos_box(:,n,nt) = xpos_box(:,n,nt)/area_box
    end do
  end do

  ! write to state vector
  do n = 1, ntstate
    if ( config%spec.eq.'co2' ) then
      do nt = 1, ndt
        states%px((n-1)*ndt*nbox+(nt-1)*nbox+1:(n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) = xpos_box(:,n,nt)
      end do
    else if ( config%spec.eq.'ghg' ) then
      states%px((n-1)*nbox+1:(n-1)*nbox+nbox_xy(ix,jy)) = xpos_box(:,n,1)
    endif
  end do

  deallocate( xpos )
  deallocate( xpos_box )


end subroutine read_bg


