!---------------------------------------------------------------------------------------
! FLEXINVERT: read_obs
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> read_obs
!!
!! Purpose:    Read observations from files and write to obsread.txt
!!
!---------------------------------------------------------------------------------------

subroutine read_obs(config, files)

  use mod_settings
  use mod_var
  use mod_dates
  use mod_strings

  implicit none

  type (config_t), intent(in)                            :: config
  type (files_t),  intent(in)                            :: files
  character(len=max_path_len), dimension(:), allocatable :: filelist
  character(len=4)                                       :: recs
  character(len=20)                                      :: species, before, sep
  character(len=200)                                     :: header, string
  integer                                                :: ierr
  integer                                                :: cnt
  integer                                                :: nr, nf, nfiles
  integer                                                :: yyyymmdd, hhmiss, hl, hh
  real(kind=8)                                           :: jdate, avetime
  real                                                   :: conc, err, measerr, lon, alt
  integer                                                :: num, reclen 

  ! list observation files
  ! ----------------------

  call system('ls '//trim(files%path_obs)//'*'//trim(files%suffix)//' | wc -l > '//trim(files%path_output)//'obsfiles.txt')
  call system('ls '//trim(files%path_obs)//' | grep '//trim(files%suffix)//' >> '//trim(files%path_output)//'obsfiles.txt')

  open(100,file=trim(files%path_output)//'obsfiles.txt',action='read',status='old',iostat=ierr)
  if ( ierr.ne. 0 ) then
    write(logid,*) 'ERROR: cannot open obsfiles.txt'
    stop
  endif
  read(100,*,iostat=ierr) nfiles
  allocate ( filelist(nfiles), stat = ierr )
  do nf = 1, nfiles
    read(100,fmt='(A)',iostat=ierr) filelist(nf)
    if (ierr.ne.0) exit
  end do
  close(100)

  ! loop over files
  ! ---------------

  ! output file for all observations
  open(200,file=trim(files%path_output)//'obsread.txt',action='write',status='old',iostat=ierr)
  if (ierr.ne.0) then
    write(logid,*) 'ERROR: cannot open obsread.txt'
    stop
  endif
  write(200,*) 'rec yyyymmdd hhmmss juldate avetime conc error'

  cnt = 0
  do nf = 1, nfiles

    ! check this file belongs to receptor in reclist
    string = filelist(nf)
    call split(string,".",before,sep)
    call split(before,"_",recs,sep)
    species = trim(before)
    reclen = len_trim(recs)

    if ( .not.( any(recname(:)(1:reclen).eq.to_upper(recs)) ) ) go to 10

    ! open input file
    open(100,file=trim(files%path_obs)//trim(filelist(nf)),action='read',status='old',iostat=ierr)
    if ( ierr.ne.0 ) then
      write(logid,*) 'WARNING: cannot open: '//trim(files%path_obs)//trim(filelist(nf))
      go to 10
    endif 
    write(logid,*) 'Reading file: '//trim(files%path_obs)//trim(filelist(nf))

    ! read header
    read (100, fmt='(A)', iostat=ierr) header

    ! station coordinates for data selection
    do nr = 1, nrec
      if ( to_upper(recs).eq.recname(nr) ) then
        lon = reclon(nr)
        alt = recalt(nr)
        go to 20
      endif
    end do
20 continue
    if ( lon.eq.0.and.alt.eq.0 ) then 
      write(logid,*) 'ERROR: read_obs: lon and alt are both zero -> check recepter list file'
      stop
    endif

    ! read data
    read_loop: do 
      ! if avetime included in obs file
      if (reclen.eq.4) then
          read(100,fmt='(A4,1X,I8,1X,I6,1X,F14.6,1X,F10.6,1X,F10.4,1X,F10.4,1X,I4)',iostat=ierr) &
              recs, yyyymmdd, hhmiss, jdate, avetime, conc, err, num
      elseif (reclen.eq.3) then
          read(100,fmt='(A3,1X,I8,1X,I6,1X,F14.6,1X,F10.6,1X,F10.4,1X,F10.4,1X,I4)',iostat=ierr) &
              recs, yyyymmdd, hhmiss, jdate, avetime, conc, err, num
      endif
     ! if avetime not included (backwards compatability with observations prepared before update)
!      avetime = 0d0
!      if (reclen.eq.4) then
!          read(100,fmt='(A4,1X,I8,1X,I6,1X,F14.6,1X,F10.4,1X,F10.4,1X,I4)',iostat=ierr) &
!              recs, yyyymmdd, hhmiss, jdate, conc, err, num
!      elseif (reclen.eq.3) then
!          read(100,fmt='(A3,1X,I8,1X,I6,1X,F14.6,1X,F10.4,1X,F10.4,1X,I4)',iostat=ierr) &
!              recs, yyyymmdd, hhmiss, jdate, conc, err, num
!      endif
      if ( ierr.ne.0 ) exit read_loop
      if ( jdate.ge.(juldatef+1d0) ) exit read_loop
      if ( jdate.lt.juldatei ) cycle read_loop
      ! select day/night for low/high alt sites
      hh = hhmiss/10000
      hl = hh + int(lon*24./360.)
      if ( hl.lt.0 ) then
        hl = hl + 24
      else if ( hl.ge.24 ) then
        hl = hl - 24
      endif
      if ( alt.le.1000. ) then
!        if ( (hl.lt.11).or.(hl.gt.17) ) cycle read_loop
      else
!        if ( (hl.gt.4).and.(hl.lt.22) ) cycle read_loop
      endif
      if ( conc.le.-999. ) cycle read_loop
      measerr = max(config%measerr, err)
      cnt = cnt + 1
      ! write to output file
      write(200,fmt='(A4,1X,I8,1X,I6,1X,F14.6,1X,F10.6,1X,F10.4,1X,F10.4)') &
              recs, yyyymmdd, hhmiss, jdate, avetime, conc, measerr
    end do read_loop

    ! close input file
    close(100)

10  continue

  end do

  ! close output file
  close(200)

  ! count observations
  nobs = nobs + cnt

end subroutine read_obs


